from profile import *
from resources import ResourceFinder
from template import *
from util import *
from traceback import print_exception
import urlparse

log = logging.getLogger("wooly")
strings = StringCatalog(__file__)

class CSRFException(Exception):
    pass

# XXX make this more definitely WidgetAttribute
class Attribute(object):
    def __init__(self, app, name):
        self.app = app
        self.name = name
        self.widget = None
        self.default = None
        self.required = True

        self.path = None

    def init(self):
        log.debug("Initializing %s", self)

        if self.widget and self.widget.path:
            self.path = ".".join((self.widget.path, self.name))
        else:
            self.path = self.name

    def seal(self):
        log.debug("Sealing %s", self)

    def validate(self, session):
        value = self.get(session)

        if value is None and self.required:
            raise Exception("%s not set" % self)

    def get(self, session):
        value = session.get(self.path)

        if value is None:
            value = self.get_default(session)

            if value is not None:
                self.set(session, value)

        return value

    def add(self, session, value, key=None):
        self.set(session, value)

    def set(self, session, value):
        # XXX don't return the value
        return session.set(self.path, value)

    def unset(self, session):
        session.unset(self.path)

    def get_default(self, session):
        return self.default

    def __repr__(self):
        return "%s('%s')" % (self.__class__.__name__, self.path)

class ListAttribute(Attribute):
    def get_default(self, session):
        return list()

    def add(self, session, value, key=None):
        self.get(session).append(value)

class Parameter(Attribute):
    def __init__(self, app, name):
        super(Parameter, self).__init__(app, name)

        self.is_collection = False
        self.is_dictionary = False

    def init(self):
        super(Parameter, self).init()

        self.widget.page.init_parameter(self)

    def marshal(self, object):
        if object == None:
            string = ""
        else:
            string = self.do_marshal(object)

        return string

    def do_marshal(self, object):
        return str(object)

    def unmarshal(self, string):
        return self.do_unmarshal(string)

    def do_unmarshal(self, string):
        return string

class Widget(object):
    def __init__(self, app, name):
        self.app = app
        self.name = name
        self.parent = None
        self.children = list()
        self.children_by_name = dict()
        self.attributes = list()
        self.parameters = list()

        self.sealed = False

        # Configuration
        self.html_class = None
        self.update_enabled = False
        self.defer_enabled = False
        self.comments_enabled = True
        self.check_login = False

        # These are computed later in the init pass
        self.ancestors = None
        self.path = None
        self.page = None
        self.frame = None

        self.__main_tmpl = WidgetTemplate(self, "html")
        self.__defer_tmpl = WidgetTemplate(self, "deferred_html")
        self.nocontent_tmpl = WidgetTemplate(self, "nocontent")

        self.my_name = self.__class__.__module__ + '.' + self.__class__.__name__

    def init(self):
        log.debug("Initializing %s" % self)

        self.init_computed_values()

        if self.page:
            self.page.init_widget(self)

        for attr in self.attributes:
            attr.init()

        for param in self.parameters:
            param.init()

        for child in self.children:
            child.init()

    def init_computed_values(self):
        ancestors = list()
        widget = self.parent

        while widget is not None:
            ancestors.append(widget)
            widget = widget.parent

        self.ancestors = tuple(ancestors)

        pelems = [x.name for x in reversed(ancestors)]
        pelems.append(self.name)
        self.path = ".".join(pelems[1:])

        self.page = self.ancestors[-1]

        for widget in ancestors:
            if isinstance(widget, Frame):
                self.frame = widget
                break

        if not self.html_class:
            self.init_html_class()

    def init_html_class(self):
        tokens = list()

        if self.page.page_html_class:
            tokens.append(self.page.page_html_class)

        for cls in self.__class__.__mro__:
            tokens.append(cls.__name__)

            if cls is Widget:
                break

        if not tokens:
            tokens.append("_")

        self.html_class = " ".join(tokens)

    def seal(self):
        assert not self.sealed
        self.sealed = True

        assert self.path is not None, self
        assert isinstance(self.page, Page), self.page

        for attr in self.attributes:
            attr.seal()

        for param in self.attributes:
            param.seal()

        for child in self.children:
            child.seal()

    def add_child(self, child):
        assert not self.sealed, self
        assert isinstance(child, Widget), (self, child)
        assert child is not self, self
        assert child not in self.children, (self, child, self.children)
        assert child.name not in self.children_by_name, child.name

        self.children.append(child)
        self.children_by_name[child.name] = child
        child.parent = self

    def remove_child(self, child):
        assert not self.sealed
        assert isinstance(child, Widget)
        assert child is not self

        self.children.remove(child)
        del self.children_by_name[child.name]
        child.parent = None

    def replace_child(self, child):
        existing = self.children_by_name[child.name]

        self.remove_child(existing)
        self.add_child(child)

    def add_attribute(self, attribute):
        assert not self.sealed
        assert isinstance(attribute, Attribute)

        self.attributes.append(attribute)
        attribute.widget = self

    def add_parameter(self, parameter):
        assert not self.sealed
        assert isinstance(parameter, Parameter)

        self.parameters.append(parameter)
        parameter.widget = self

    def get_ancestor(self, name):
        for anc in self.ancestors:
            if anc.name == name:
                return anc

    @classmethod
    def get_module_strings(cls):
        module = sys.modules[cls.__module__]
        return module.__dict__.get("strings")

    def get_string(self, key):
        for cls in self.__class__.__mro__:
            if cls is object:
                return

            str = None
            strs = cls.get_module_strings()

            if strs:
                str = strs.get(cls.__name__ + "." + key)

                if str:
                    return str

            str = cls.__dict__.get(key)

            if str:
                return str

    def show(self, session):
        assert self.parent

        self.parent.show_child(session, self)
        self.parent.show(session)

        return self

    def show_child(self, session, child):
        pass

    def save_parameters(self, session, params):
        params.extend(self.parameters)

        for child in self.children:
            child.save_parameters(session, params)

    def redirect_on_not_authorized(self, session):
        # Give an opportunity to redirect to a valid page
        # when a user is not authorized for this widget.
        mainpage = self.app.mainpage_cb(session)
        if mainpage in self.app.pages_by_name:
            log.debug("Not authorized, redirecting to %s", mainpage)
            page = self.app.pages_by_name[mainpage]
            sess = Session(page)
            if hasattr(self, "redirect"):
                self.redirect.set(session, sess.marshal())
            elif hasattr(self, "page"):
                self.page.redirect.set(session, sess.marshal())
        
    def authorized(self, session):
        return session.login_current()

    def redirect_to_login(self, session):
        # Redirect to the login page when not logged in.
        # For some page types (like Update pages) this
        # is accomplished differently so allow an overload...
        page = self.app.login_page
        sess = Session(page)
        page.origin.set(sess, session.marshal())
        self.redirect.set(session, sess.marshal())      

    def process(self, session, *args):
        if self.check_login and not self.authorized(session):
            self.redirect_to_login(session)
            return

        class Reason(object):
            pass

        reason = Reason()
        if not self.app.authorize_cb(session, self, reason):
            # It's possible that a widget is not authorized
            # because its module is not loaded, in which case
            # it should simply not draw (no redirect needed).
            # So check the reason before we redirect.
            if reason.reason == "GROUP_NOT_AUTHORIZED":
                self.redirect_on_not_authorized(session)
        else:
            if self.update_enabled:
                self.page.enable_update(session, self)

            if self.defer_enabled:
                self.page.enable_defer(session, self)

            if self.app.debug:
                profile = self.page.profile.get(session)

                call = ProcessCall(profile, self, args)
                call.do(session)
            else:
                self.do_process(session, *args)

    def do_process(self, session, *args):
        for child in self.children:
            child.process(session)

    def render(self, session, *args):
        if not self.app.authorize_cb(session, self):
            return ""
        if self.app.debug:
            profile = self.page.profile.get(session)

            call = RenderCall(profile, self, args)
            string = call.do(session)
        else:
            string = self.do_render(session, *args)

            if string is None:
                string = ""

        return string

    def do_render(self, session, *args):
        if self.defer_enabled and not getattr(session, "background", None):
            writer = Writer()
            self.__defer_tmpl.render(writer, session, *args)
            return writer.to_string()
        else:
            writer = Writer()
            self.__main_tmpl.render(writer, session, *args)
            return writer.to_string()

    def render_id(self, session, *args):
        return self.path

    def render_class(self, session, *args):
        return self.html_class

    def render_href(self, session, *args):
        return session.marshal()

    def render_title(self, session, *args):
        return None

    def render_resource_prefix(self, session, *args):
        return self.page.get_resource_prefix(session)

    def render_content(self, session, *args):
        writer = Writer()

        for child in self.children:
           writer.write(child.render(session))

        return writer.to_string()

    def render_deferred_content(self, session, *args):
        return None

    def prt(self):
        print self

        for key, value in sorted(self.__dict__.items()):
            print "    %-30s %s" % (key + ":", value)

    def __repr__(self):
        return "%s.%s('%s')" % \
            (self.__class__.__module__, self.__class__.__name__, self.path)
            
    def render_error_msg(self, session, *args):
        return "There is no content to display"

    def set_redirect(self, session, frame, message):
        nsession = session.branch()
        frame.view.show(nsession)
        url = nsession.marshal()
        self.page.redirect.set(session, url)
        if message:
            session.add_notice(Notice(message))

    def get_redirect(self, session):
        return self.page.redirect.get(session)

    def validate_on_unmarshal(self, param, value, session):
        return True

class Frame(Widget):
    def show(self, session):
        super(Frame, self).show(session)

        return self.page.set_frame(session, self)

    def save_parameters(self, session, params):
        frame = self.page.get_frame(session)

        if self is frame or self in frame.ancestors:
            super(Frame, self).save_parameters(session, params)

class Page(Frame):
    xml_content_type = "text/xml"
    html_content_type = "text/html"
    xhtml_content_type = "application/xhtml+xml"
    xml_1_0_declaration = """<?xml version="1.0" encoding="UTF-8"?>"""
    xhtml_1_1_doctype = """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">"""
    html_doctype = """<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">"""
    xhtml_namespace = "http://www.w3.org/1999/xhtml"

    def __init__(self, app, name):
        super(Page, self).__init__(app, name)

        self.page_html_class = None

        self.page_widgets = list()
        self.page_widgets_by_path = dict()

        self.page_parameters = list()
        self.page_parameters_by_path = dict()

        self.redirect = Attribute(app, "redirect")
        self.add_attribute(self.redirect)

        self.allow_object_not_found = Attribute(app, "allow_object_not_found")
        self.allow_object_not_found.required = False
        self.add_attribute(self.allow_object_not_found)

        self.error = Attribute(app, "error")
        self.add_attribute(self.error)

        self.profile = self.ProfileAttribute(app, "profile")
        self.add_attribute(self.profile)

        # All pages should only be processed if the user
        # is logged in, except for the login page, resource, and css.
        self.check_login = True

    def init_computed_values(self):
        self.ancestors = ()
        self.path = ""
        self.page = self
        self.frame = self

        if not self.html_class:
            self.init_html_class()

    def init_widget(self, widget):
        assert not self.sealed
        assert isinstance(widget, Widget)

        self.page_widgets.append(widget)
        self.page_widgets_by_path[widget.path] = widget

    def init_parameter(self, param):
        assert not self.sealed
        assert isinstance(param, Parameter)
        assert param.path != Session.csrf_tag_name

        self.page_parameters.append(param)
        self.page_parameters_by_path[param.path] = param

    def get_page_parameter_by_path(self, path):
        param = self.page_parameters_by_path.get(path)

        # check for partial match for dictparameters
        if not param:
            if DictParameter.sep() in path:
                dict_key = path.split(DictParameter.sep(), 1)[0]
                param = self.page_parameters_by_path.get(dict_key)

        return param

    def get_page_parameters(self, session):
        return self.page_parameters

    def get_last_modified(self, session):
        return None

    def get_content_type(self, session):
        raise Exception("Not implemented")

    def get_extra_headers(self, session):
        return []

    def get_cache_control(self, session):
        return None

    def get_resource_prefix(self, session):
        return "resource?name="

    def show(self, session):
        pass

    def save_session(self, session):
        pass

    def enable_update(self, session, widget):
        pass

    def service(self, session):
        self.process(session)

        redirect = self.redirect.get(session)
        if redirect:
            raise PageRedirect()

        rend = self.render(session)

        redirect = self.redirect.get(session)
        if redirect:
            raise PageRedirect()

        return rend

    def validate_redirect(self, session, url, env):

        # Sigh, urlparse returns a different result in Python 2.4 and
        # Python 2.6 so we have to index ourselves.
        SCHEME   = 0
        NETLOC   = 1
        PATH     = 2
        PARAMS   = 3
        QUERY    = 4
        FRAGMENT = 5

        def netloc(env):
            if 'HTTP_HOST' in env:
                return env['HTTP_HOST']
            return env['SERVER_NAME']

        # Only allow redirects to our own pages
        if type(url) in (str, unicode):
            res = urlparse.urlparse(url)
            # Path will be an index into our dictionary of pages
            p = res[PATH]
            if p.startswith("/"):
                p = p[1:]

            if not p in self.app.pages_by_name or \
               (res[SCHEME] and res[SCHEME] != env['wsgi.url_scheme']) or \
               (res[NETLOC] and res[NETLOC] != netloc(env)):
                session.add_notice(Notice("Invalid redirect, " \
                                          "returning to main page"))
                url = self.lookup_mainpage(session)
        else:
            url = self.lookup_mainpage(session)
        return url

    def lookup_mainpage(self, session):
        if self.app.authorizator.is_enforcing():
            mainpage = self.app.authorizator.find_mainpage(session)
        else:
            mainpage = "index.html"
        return mainpage

    def redirect_on_exception(self, session):
        return None

    def service_error(self, session):
        # Well, we can have other handled exceptions pop up before
        # we actually get to the render so we want to record here
        # what the root cause was.
        cls, value, traceback = sys.exc_info()
        self.error.set(session, PageError(self, session, 
                                          cls, value, traceback))

        return self.render(session)

    def render_id(self, session, *args):
        return self.name

    def render_content(self, session):
        if self.error.get(session):
            return self.render_error(session)

        return super(Page, self).render_content(session)

    def render_error(self, session):
        return self.error.get(session).render()

    class ProfileAttribute(Attribute):
        def get_default(self, session):
            return PageProfile(self)

class PageRedirect(Exception):
    pass

class UpdateRedirect(Exception):
    def __init__(self, url):
        self.url = url

class PageError(object):
    def __init__(self, page, session, cls, value, traceback):
        self.page = page
        self.session = session
        self.cls = cls
        self.value = value
        self.traceback = traceback

    def render(self):
        writer = Writer()
        writer.write("APPLICATION ERROR\n\n")

        print_exception(self.cls, self.value, self.traceback, None, writer)

        writer.write("\n")

        profile = self.page.profile.get(self.session)

        if profile:
            writer.write("Widget trace:\n\n")
            profile.print_stack_trace(writer)
            writer.write("\n")

        self.print_messages(writer)
        self.print_session(writer)

        env = self.session.request_environment

        if env:
            self.print_url_vars(env["QUERY_STRING"], writer)
            self.print_environment(env, writer)

        return "<pre>%s</pre>" % xml_escape(writer.to_string())

    def print_messages(self, writer):
        writer.write("Messages:\n\n")

        for message in self.session.messages:
            writer.write("  %s\n" % message)

        writer.write("\n")

    def print_session(self, writer):
        writer.write("Session:\n\n")

        for path in sorted(self.session.values_by_path):
            value = self.session.values_by_path[path]

            writer.write("  %-30s  %s\n" % (path, value))

        writer.write("\n")

    def print_url_vars(self, query, writer):
        writer.write("URL variables:\n\n")

        if query:
            vars = query.split(";")

            for var in sorted(vars):
                try:
                    key, value = var.split("=")
                except ValueError:
                    continue

                writer.write("  %-30s  %s\n" % (key, value))

        writer.write("\n")

    def print_environment(self, env, writer):
        writer.write("Environment:\n\n")

        for key in sorted(env):
            value = env[key]

            writer.write("  %-30s  %s\n" % (key, value))

        writer.write("\n")

from parameters import DictParameter

class Application(object):
    def __init__(self):
        self.pages = list()
        self.pages_by_name = dict()
        self.page_links = list()

        self.resource_page = ResourcePage(self, "resource")
        self.add_page(self.resource_page)

        self.finder = ResourceFinder()

        self.debug = False

    def init(self):
        for page in self.pages:
            page.init()

        for page in self.pages:
            page.seal()

    def add_page(self, page, add_to_link_set=False):
        if page.parent:
            raise Exception("Page '%s' is not a root widget" % page.name)

        if page.name in self.pages_by_name:
            raise Exception \
                ("A page called '%s' has already been added" % page.name)

        self.pages.append(page)
        self.pages_by_name[page.name] = page
        if add_to_link_set:
            self.page_links.append(page)

    def set_default_page(self, page):
        self.pages_by_name[""] = page

    def add_resource_dir(self, dir):
        self.finder.add_dir(dir)

    def get_resource(self, name):
        return self.finder.find(name)

    def __repr__(self):
        return self.__class__.__name__

class Session(object):
    http_date_gmt = "%a, %d %b %Y %H:%M:%S GMT"

    csrf_tag_name = "csrf"

    def __init__(self, page):
        assert isinstance(page, Page)

        self.app = page.app
        self.page = page

        self.client_session = None

        self.trunk = None

        self.values_by_path = dict()
        self.cookies_by_name = dict() # name => (newly set?, value, expires)

        self.request_environment = None

        self.messages = list()

        # This will be the value of the csrf tag if
        # it was passed in the request
        self.csrf_value = None
        self.post = False

    def login_current(self, login=None):
        if not login:
            login = self.client_session.attributes.get("login_session")
        if login:
            when = datetime.now() - timedelta(hours=24)
            if login.created > when:
                return True
        return False

    def check_csrf(self, page):
       if self.client_session.get_csrf() != self.csrf_value:
           # For a current login, log that there may have been
           # a possible CSRF attempt.  If the client is not
           # logged in then this is probably due to a stale login
           # form instead.
           if self.login_current():
               log.info("Possible CSRF attempt using %s" % self)

           # If we are currently on the login page,redirect 
           # back to the login page instead of the mainpage
           # with a notice explaining the need to login in again.
           # If we go to the mainpage we will lose the warning 
           # because the 'login_invalid' parameter is only a page
           # parameter for the login_page and will not be rendered
           # in the url for other pages.
           if page == self.app.login_page:
               sess = Session(page)
               page.login_form.login_invalid.set(sess, "form")
           else:
               mainpage = None
               if self.app.authorizator.is_enforcing():
                   try:
                       p = self.app.authorizator.find_mainpage(self)
                       mainpage = self.app.pages_by_name[p]
                   except:
                       pass
               if mainpage is None:
                   mainpage = self.app.pages_by_name["index.html"]
               sess = Session(mainpage)
           page.redirect.set(self, sess.marshal())
           raise PageRedirect()
    
    def branch(self):
        session = Session(self.page)
        session.trunk = self

        return session

    def log(self, message):
        self.messages.append(message)

    def get_cookie(self, name):
        try:
            return self.cookies_by_name[name][1]
        except KeyError:
            pass

    def set_cookie(self, name, value, expires=None):
        self.cookies_by_name[name] = (True, value, expires)

    def expire_cookie(self, name):
        dawn_of_time = time.gmtime(1)[0:6]
        self.set_cookie(name, "", datetime(*dawn_of_time))

    def get(self, key):
        if key in self.values_by_path:
            value = self.values_by_path[key]
        elif self.trunk:
            value = self.trunk.get(key)
        else:
            value = None

        return value

    def set(self, key, value):
        #if key.startswith("server.vhost.queue.edit.durable"):
        #    raise Exception()

        self.values_by_path[key] = value

        return value

    def unset(self, key):
        if key in self.values_by_path:
            del self.values_by_path[key]

    def marshal(self):
        page = self.page.name
        vars = self.marshal_url_vars()

        if vars:
            url = page + "?" + vars
        else:
            url = page

        if self.app.debug: 
            profile = self.page.profile.get(self)
            profile.urls.append(url)

        return url

    def marshal_url_vars(self, separator=";"):
        params = self.page.get_page_parameters(self)
        vars = list()

        values_by_path = self.gather_values()

        for param in params:
            key = param.path
            value = values_by_path.get(key)

            if value is not None:
                if param.is_collection:
                    for item in value:
                        sitem = quote(param.marshal(item))
                        vars.append("%s=%s" % (key, sitem))
                elif value != param.get_default(self):
                    svalue = quote(param.marshal(value))
                    vars.append("%s=%s" % (key, svalue))

        return separator.join(vars)
            
    def gather_values(self):
        if self.trunk is None:
            return self.values_by_path
        else:
            values_by_path = dict()

            values_by_path.update(self.trunk.gather_values())
            values_by_path.update(self.values_by_path)

            return values_by_path

    def unmarshal(cls, app, string):
        elems = string.split("?")

        try:
            name = elems[0]

            if name.startswith("/"):
                name = name[1:]

            page = app.pages_by_name[name]
        except KeyError:
            raise Exception("Page '%s' not found" % name)

        session = Session(page)

        try:
            session.unmarshal_url_vars(elems[1])
        except IndexError:
            pass

        return session

    unmarshal = classmethod(unmarshal)

    def unmarshal_url_vars(self, string, separator=";"):
        from wooly.forms import FormInput
        vars = string.split(separator)

        for var in vars:
            try:
                skey, svalue = var.split("=")
            except ValueError:
                continue

            key = unquote(skey)
            value = unquote_plus(svalue)

            param = self.page.get_page_parameter_by_path(key)
            if param:
                if not param.widget.validate_on_unmarshal(param, value, self):
                    # Don't try setting a form input anywhere other than
                    # the file stream of a POST request!
                    raise Exception("Illegal attempt to set parameter %s" % param)
                param.add(self, param.unmarshal(value), key)

            elif key == self.csrf_tag_name:
                self.csrf_value = value

    def marshal_cookies(self, secure_cookies):
        """
        Produce a list of strings representing cookies newly set on
        this session.

        The strings are formatted to serve as HTTP Set-Cookie header
        values.
        """

        headers = list()

        for name, record in self.cookies_by_name.iteritems():
            new, value, expires = record

            if new:
                crumbs = list()
                crumbs.append("%s=%s" % (name, value))

                if expires:
                    when = expires.strftime(self.http_date_gmt)
                    crumbs.append("expires=%s" % when)

                if secure_cookies:
                    crumbs.append("secure")

                headers.append("; ".join(crumbs))

        return headers

    def unmarshal_cookies(self, string):
        """
        Parse and load cookies from string.

        The string must be formatted as dictated by the HTTP Cookie
        header.
        """

        for crumb in string.split(";"):
            name, value = crumb.split("=", 1)
            self.cookies_by_name[name.strip()] = (False, value.strip(), None)

    def add_notice(self, notice):
        try:
            n = self.client_session.attributes["login_session"].notifications
            n.append(notice)
        except:
            pass
        
    def add_unique_notice(self, notice):
        try:
            notifications = self.client_session.attributes["login_session"].notifications
            for n in notifications:
                if not n.dismissed and n.message == notice.message:
                    # this would be a duplicate message, we can bail out
                    return False
            notifications.append(notice)
        except:
            pass
        return True

    def get_notifications(self):
        try:
            return self.client_session.attributes["login_session"].notifications
        except:
           return []

    def get_user(self):
        return self.client_session.attributes["login_session"].user

    def __repr__(self):
        return "%s(trunk=%s,app=%s,id=%i)" % \
            (self.__class__.__name__, self.trunk, self.app, id(self))

class Notice(object):
    def __init__(self, message, preamble="Notice", capitalize=True):
        self.dismissed = False
        if capitalize:
            message = message.capitalize()
        self.message = message
        self.preamble = preamble
        self.timestamp = datetime.now()  #may want a superclass since this is duplicated in TaskInvocation

    def get_message(self, session):
        timestamp = "%s-%02d-%02d %02d:%02d:%02d" % (self.timestamp.year, 
                                                     self.timestamp.month, 
                                                     self.timestamp.day, 
                                                     self.timestamp.hour, 
                                                     self.timestamp.minute, 
                                                     self.timestamp.second)
        return "<span class='notification_time'>%s</span> %s: %s" % (timestamp, 
                                                                   self.preamble,
                                                                   self.message)

class SessionAttribute(object):
    def __init__(self, widget, name):
        assert hasattr(widget, "app"), widget
        
        self.widget = widget
        self.name = name
        self.default = None

        self.key = (self.widget, self.name)

    def do_get(self, session):
        return copy(self.default)

    def get(self, session):
        value = session.get(self.key)

        if value is None:
            value = self.do_get(session)

            if value is not None:
                self.set(session, value)

        return value

    def add(self, session, value):
        self.set(session, value)

    def set(self, session, value):
        session.set(self.key, value)

from pages import ResourcePage
